﻿#include <stdlib.h>
#include <string.h>
#include <security/pam_appl.h>
#include <security/pam_modules.h>

#include <vector>
#include <string>
#include <syslog.h>

// getuid
#include <unistd.h>
//getpwnam
#include <sys/types.h>
#include <pwd.h>

//forward 
std::vector<unsigned char> ModhexDecode(const std::string& otp_entry);

#include "ParseOTP.h"

typedef unsigned char BYTE;
typedef ParsedOTPDataT<std::string, std::string, std::vector<BYTE>> ParsedOTPData;
bool ParseOTP(const std::string& username, const std::string& password, ParsedOTPData& res, unsigned short expectedYubikeyOtpChars, unsigned short expectedOathOtpChars)
{
	return ParseOTP<std::string, char, std::string, std::string, std::vector<BYTE>, std::vector<BYTE> >(username, password, res, expectedYubikeyOtpChars, expectedOathOtpChars);
}
bool ParseOTP(const std::string& username, const void* authtok, ParsedOTPData& res)
{
	return ParseOTP(username, std::string((char*)authtok), res, EXPECTED_YUBIKEY_OTP_CHARACTERS, EXPECTED_OATH_OTP_CHARACTERS);
}

/* expected hook */
PAM_EXTERN int pam_sm_setcred(pam_handle_t* pamh, int flags, int argc, const char** argv) {
	return PAM_IGNORE;
}

PAM_EXTERN int pam_sm_acct_mgmt(pam_handle_t* pamh, int flags, int argc, const char** argv) {
	return PAM_IGNORE;
}

void FreeResponseArray(pam_response* response_array)
{
	// assuming an array of 1 since we only call it with 1 message
	if (NULL != response_array->resp)
	{
		free(response_array->resp);
	}
	free(response_array);
}

int LogIfError(int err, const char* fn)
{
	if (PAM_SUCCESS != err)
	{
		syslog(LOG_ERR, "Error in function %s: %d", fn, err);
	}
	return err;
}
/* expected hook, this is where custom stuff happens */
PAM_EXTERN int pam_sm_authenticate(pam_handle_t* pamh, int flags, int argc, const char** argv) {
	openlog("pam_authlite", LOG_NDELAY, LOG_AUTH);

	const char* pUsername;
	if (PAM_SUCCESS == LogIfError(pam_get_user(pamh, &pUsername, "Username: "), "pam_get_user"))
	{
		syslog(LOG_INFO, "Version %s; uid %d; Incoming username: %s", VERSION, getuid(), pUsername);
	}
	std::string domain, username, sep;
	username = std::string(pUsername);
	std::list<std::string> duinput;
	duinput = SplitString(username, '\\');

	if (duinput.size() > 1)
	{
		domain = *(duinput.begin()); //[0]
		sep = "\\";
		username = *(++duinput.begin()); //[1]
	}

	const void* authtok = NULL;
	pam_get_item(pamh, PAM_AUTHTOK, &authtok);
	if(NULL != authtok)
	{
		DEBUGLOG("Password already available");
	}
	else
	{
		DEBUGLOG("No password available, requesting");
		const struct pam_conv* conv;
		if (PAM_SUCCESS != LogIfError( pam_get_item(pamh, PAM_CONV, (const void**)&conv), "pam_get_item(PAM_CONV)") || NULL == conv)
		{
			return PAM_IGNORE;
		}
		struct pam_message theMessage;
		const struct pam_message* msgPtr = &theMessage;
		theMessage.msg = "Password:";
		theMessage.msg_style = PAM_PROMPT_ECHO_OFF;

		struct pam_response* response_array = NULL;
		if (PAM_SUCCESS != LogIfError(conv->conv(1, &msgPtr, &response_array, conv->appdata_ptr), "conv->conv") || NULL == response_array)
		{
			return PAM_IGNORE;
		}

		DEBUGLOG("Received password");

		if (PAM_SUCCESS != LogIfError(pam_set_item(pamh, PAM_AUTHTOK, response_array->resp), "pam_set_item(PAM_AUTHTOK)"))
		{
			FreeResponseArray(response_array);
			return PAM_IGNORE;
		}
		FreeResponseArray(response_array);
		response_array = NULL;

		pam_get_item(pamh, PAM_AUTHTOK, &authtok);
	}

	ParsedOTPData parsed;
	if (ParseOTP(username, authtok, parsed))
	{
		NORMALLOG("Parsed OTP of type %d. Domain: %s Username: %s OTP: %s", parsed.otp_type, domain.c_str(), parsed.username.c_str(), 
			parsed.otp_string_short.empty()? parsed.otp_string.c_str() : parsed.otp_string_short.c_str());

		std::string fmt = "";
		if(domain.size() > 0)
		{	
			fmt += domain;
			fmt += sep;
		}
		fmt += parsed.otp_string;

#ifdef USE_GETPWNAM
		SUPERLOG("getpwnam('%s')", fmt.c_str());
		getpwnam(fmt.c_str());
#endif
		fmt = "";
		if(domain.size() > 0)
		{	
			fmt += domain;
			fmt += sep;
		}
		fmt += parsed.username;
		SUPERLOG("pam_set_item(PAM_USER, '%s')", fmt.c_str());
		LogIfError(pam_set_item(pamh, PAM_USER, fmt.c_str()), "pam_set_item(PAM_USER)");
		SUPERLOG("pam_set_item(PAM_AUTHTOK, '%s')", parsed.password.c_str());
		LogIfError(pam_set_item(pamh, PAM_AUTHTOK, parsed.password.c_str()), "pam_set_item(PAM_AUTHTOK)(2)");
	}
	else
	{
		NORMALLOG("No OTP found. Username: %s", username.c_str());
	}
	

	closelog();

	return PAM_IGNORE;
}
